var us =require('underscore'),
    async = require('async'),
    common = require('./common'),
    setReady = common.setReady,
    sendError = common.sendError;

function makeActivityMessagePretty(message) {

    var pretty = {},
        keys,
        firstMongoMessage;

    // Is it a Mongo-style error?
    if (
        us.isString(message.level) &&
        us.isString(message.message)) {

        if (message.message == 'Validation failed' && message.body) {

            keys = us.keys(message.body);
            firstMongoMessage = message.body[keys[0]];

            if (firstMongoMessage.type == 'required') {
                pretty.message = "A value must be supplied for '" + firstMongoMessage.path + "'";
            }
            else {
                pretty.message = "The value of '" + firstMongoMessage.value + "' for " + firstMongoMessage.message;
            }
        }
        else {
            pretty.message = message.message;
        }

    }
    else {
        pretty = message.mesage;
    }

    return pretty;
}

function addHumanReadableMessages(summary) {

    var flowMessage, activityMessage, i, j;
    summary.forHumans = [];

    if (us.has(summary, 'body')) {
        for ( i=0; i<summary.body.length; i++) {
            flowMessage = summary.body[i];

            if (us.has(flowMessage, 'body')) {
                for (j=0; j<flowMessage.body.length; j++ ) {
                    activityMessage = flowMessage.body[j];
                    summary.forHumans.push(makeActivityMessagePretty(activityMessage));
                }
            }
        }
    }

    return summary;

}

function addImmediatePendingActivity(summary, activityCache, flowDoc) {
    var i,
        activity,
        pendingActivities = [],
        worthAdding = false,
        ipa;

    for (i=0;i<flowDoc.activities.length;i++) {
        activity = flowDoc.activities[i];
        if (activity.state == 'pending') {
            pendingActivities.push(activity);
        }

        if (pendingActivities.length==1) {
            var activityType = pendingActivities[0].activityType;
            if (us.has(activityCache, activityType)) {
                worthAdding = true;
                ipa = pendingActivities[0].activityId;
            }
        }
    }

    if (worthAdding) {
        summary.immediatePendingActivityId = ipa;
    }

}

function getFlowSummary(flowDoc, activityCache) {
    var summary = {
            flowId: flowDoc._id
        },
        activity,
        finish,
        errors = [],
        alreadyRunningActivities = [],
        readyCount = 0,
        pendingCount = 0,
        failedCount = 0,
        dirtyCount = 0,
        context;

    finish = flowDoc.findActivity('finish');

    if (finish.state != 'finished') {

        // Not finished: so what we got?
        // -----------------------------

        for (var i=0;i<flowDoc.activities.length;i++) {

            activity = flowDoc.activities[i];

            switch (activity.state) {

                case 'ready':
                    readyCount += 1;
                    break;

                case 'dirty':
                    dirtyCount += 1;
                    errors.push({
                        name: 'dirtyActivity',
                        message: "Activity '" + activity.activityId + "' is marked as dirty",
                        body: activity.messages
                    });
                    break;

                case 'pending':
                    pendingCount += 1;
                    break;

                case 'running':
                    alreadyRunningActivities.push(activity.activityId);
                    break;

                case 'failed':
                    failedCount += 1;
                    errors.push({
                        name: 'badActivity',
                        message: "Activity '" + activity.activityId + "' is marked as failed",
                        body: activity.messages
                    });
                    break;
            }
        }

        if (failedCount > 0 ) {

            // Activities are in a failed state
            // --------------------------------

            summary.flowState = 'failed';
            summary.message = 'Unable to progress flow as ';
            if (failedCount == 1) {
                summary.message += "1 activity has failed";
            }
            else {
                summary.message += failedCount + " activities have failed";
            }
            summary.body = errors;

        }
        else if (alreadyRunningActivities.length>0) {


            // Activities are still in a running state
            // ---------------------------------------
            summary.flowState = 'running';
            summary.message = 'Unable to progress flow as ';
            if (alreadyRunningActivities.length==1) {
                summary.message += "1 activity is still running";
            }
            else {
                summary.message += alreadyRunningActivities.length + " activities are still running";
            }

        }


        else if (readyCount==0 && pendingCount==0) {

            // Not finished, but no activities are in a state to continue
            // ----------------------------------------------------------

            summary.flowState = 'lost';
            summary.message = "Flow cannot continue as there are no 'ready' or 'pending' activities";
        }

        else if ( readyCount>0 && alreadyRunningActivities.length == 0 && failedCount ==0 ) {

            // There are ready activities, so run the lot again
            // ------------------------------------------------
            summary.flowState = 'runAgain';
        }

        else if (dirtyCount > 0 ) {

            // Dirty activities
            // ----------------

            summary.flowState = 'dirty';
            summary.message = 'Unable to complete flow as ';
            if (dirtyCount == 1) {
                summary.message += "1 activity is dirty";
            }
            else {
                summary.message += dirtyCount + " activities are dirty";
            }
            summary.body = errors;

        }

        else if (readyCount ==0 && pendingCount>0) {
            summary.flowState = 'pending';

            if (pendingCount == 1) {
                summary.message = 'There is currently 1 activity pending';
            }
            else {
                summary.message = 'There are currently ' + pendingCount + ' activities pending';
            }
            addImmediatePendingActivity(summary, activityCache, flowDoc);
        }

    }
    else {

        summary.flowState = 'finished';
        summary.message = 'Flow has reached finished activity'
    }

    addHumanReadableMessages(summary);
    return summary;
}


function runActivity(context, callback) {

    var activity,
        activityType = context.activity.activityType;

    if (us.has(context.roadie.activities, activityType)) {
        activity = context.roadie.activities[activityType];
        activity.runFunction(
            context.activity,
            context,

            // Activity has completed
            function (err, result) {

                if (err) {
                    callback(err);
                }
                else {

                    // Everything is OK, so if our activity has just finished, look
                    // to tidy-up and set all target activities from here to 'ready'.

                    result.lastCompleted = Date.now();

                    setReady(result.state, result, context.flowDoc.activities);

                    callback(null);
                }

            }
        );
    }
    else {
        context.activity.state = 'failed';
        context.activity.messages.push(
            {
                level: 'error',
                name: 'unknownActivity',
                message: "Unable to find activity type '" + activityType + "'",
                body: null
            }
        );
        callback(null);
    }


}


var progressFlow;
module.exports = progressFlow = function progressFlow(roadie, firstInstance, flowDoc, templateId, params, username, res) {

    var i,
        start = flowDoc.findActivity('start'),
        contextsToRun = [],
        activity,
        now = Date.now(),
        failedCount = 0,
        alreadyRunningCount = 0,
        summary,
        runAgain,
        status;

    if (start) {

        if (start.state == 'pristine') {
            start.state = 'ready';
        }


        // * Turn all "ready" activities to running.
        // * Reset all dirty activities if this is the first pass

        // But also keep an eye open for already running activities (which will block any attempt to run
        // in parallel, and any activities which have failed, which also causes a halt.

        for (i=0;i<flowDoc.activities.length;i++) {
            activity = flowDoc.activities[i];

            switch (activity.state) {
                case 'ready':
                    activity.state = 'running';
                    activity.lastStarted = now;
                    break;

                case 'running':
                    alreadyRunningCount += 1;
                    break;

                case 'failed':
                    failedCount += 1;
                    break;

                case 'dirty':
                    if (firstInstance) {
                        activity.state = 'pristine';
                        activity.messages = [];
                        break;
                    }
            }
        }

        if (failedCount>0) {
            sendError(
                flowDoc._id,
                'alreadyFailed',
                'Flow ' + flowDoc._id + ' has ' + failedCount + ' failed activities',
                null,
                409,
                res
            );
        }
        else if (alreadyRunningCount>0) {
            sendError(
                flowDoc._id,
                'alreadyRunning',
                'Flow ' + flowDoc._id + ' has ' + alreadyRunningCount + ' running activities',
                null,
                409,
                res
            );
        }
        else {

            // All is well to start running those activities...

            flowDoc.markModified('activities');
            flowDoc.save(function (err, updatedDoc) {
                if (err) {
                    sendError(
                        flowDoc._id,
                        'runningStatusFail',
                        "Unable to update activities to 'running' status",
                        err,
                        500,
                        res
                    );
                }
                else {
                    for (i=0;i<updatedDoc.activities.length;i++) {
                        activity = updatedDoc.activities[i];
                        if (activity.state == 'running') {
                            contextsToRun.push({
                                id: activity.activityId,
                                roadie: roadie,
                                flowDoc: updatedDoc,
                                activity: activity,
                                res: res,
                                username: username
                            });
                        }
                    }

                    async.each(contextsToRun, runActivity, function (err) {

                        if (err) {

                            var body = err.body;
                            if (!body) {
                                body = err.error;
                            }

                            sendError(
                                updatedDoc._id,
                                err.name,
                                err.message,
                                body,
                                err.status,
                                res
                            );
                        }
                        else {

                            // Produce a summary of the activities as they stand now
                            // (used to return a decent response and also mark the
                            // overall state of the flow...
                            // -----------------------------------------------------
                            summary = getFlowSummary(updatedDoc, roadie.activities);
                            if (summary.flowState == 'runAgain') {
                                runAgain = true;
                                summary.flowState = 'running';
                            }
                            else {
                                runAgain = false;
                            }


                            // Save doc again, now that all the activities have completed and had their way
                            // ----------------------------------------------------------------------------


                            updatedDoc.state = summary.flowState;
                            updatedDoc.markModified('activities');
                            updatedDoc.save(function (err, completedDoc) {

                                if (err) {
                                    sendError(
                                        updatedDoc._id,
                                        'activityUpdateFail',
                                        "Unable to update completed activities",
                                        err,
                                        500,
                                        res
                                    );
                                }
                                else {
                                    if (runAgain) {
                                        progressFlow(roadie, false, completedDoc, templateId, params, username, res);
                                    }
                                    else {

                                        // All done, 'finished' activity would have archived
                                        // so delete this flow
                                        if (summary.flowState=='finished') {

                                            completedDoc.remove(function(err) {

                                                if (err) {
                                                    sendError(
                                                        updatedDoc._id,
                                                        'flowRemoveFail',
                                                        "Unable to remove completed flow",
                                                        err,
                                                        500,
                                                        res
                                                    );
                                                }
                                                else {
                                                    if (res) {
                                                        res.status(200).send(summary);
                                                    }
                                                }
                                            });

                                        }
                                        else {

                                            if (res) {
                                                if (summary.flowState == 'failed' ||
                                                    summary.flowState=='lost' ||
                                                    summary.flowState =='dirty') {
                                                    status = 500;
                                                }
                                                else {
                                                    status = 200;
                                                }
                                                res.status(status).send(summary);
                                            }

                                        }

                                    }
                                }
                            });
                        }
                    });
                }

            });
        }
    }
    else {
        sendError(
            flowDoc._id,
            'noStart',
            'Flow ' + flowDoc._id + ' has no start activity',
            flowDoc,
            404,
            res
        );
    }
};